import asyncio
import logging
import shelve
from collections import defaultdict
from typing import Union

import maufbapi.types
from maufbapi import AndroidAPI, AndroidState
from maufbapi.http import InvalidAccessToken
from maufbapi.types import mqtt as mqtt_t
from maufbapi.types.graphql import Participant
from mautrix.util.proxy import ProxyHandler
from slidge import BaseSession, FormField, SearchResult
from slixmpp.exceptions import XMPPError

from . import config
from .client import AndroidMQTT
from .contact import Contact, Roster
from .group import MUC, Bookmarks
from .util import FacebookMessage, Messages, get_shelf_path

Recipient = Union[Contact, MUC]


class Session(BaseSession[str, Recipient]):
    contacts: Roster
    bookmarks: Bookmarks

    mqtt: AndroidMQTT
    api: AndroidAPI
    me: maufbapi.types.graphql.OwnInfo
    my_id: int

    def __init__(self, user):
        super().__init__(user)

        # keys = "offline thread ID"
        self.ack_futures = dict[int, asyncio.Future[FacebookMessage]]()

        # keys = "facebook message id"
        self.reaction_futures = dict[str, asyncio.Future]()
        self.unsend_futures = dict[str, asyncio.Future]()

        # keys = "contact ID"
        self.sent_messages = defaultdict[int, Messages](Messages)
        self.received_messages = defaultdict[int, Messages](Messages)

    async def login(self):
        shelf: shelve.Shelf[AndroidState]
        with shelve.open(get_shelf_path(self.user_jid.bare)) as shelf:
            s = shelf["state"]
        x = ProxyHandler(None)
        self.api = AndroidAPI(state=s, proxy_handler=x)
        self.mqtt = AndroidMQTT(
            self, self.api.state, proxy_handler=self.api.proxy_handler
        )
        try:
            self.me = await self.api.get_self()
        except InvalidAccessToken:
            await self.xmpp.unregister_user(self.user)
            raise XMPPError(
                "not-authorized",
                "Your facebook access token has expired. "
                "Please re-register to the gateway.",
            )
        self.my_id = int(self.me.id)  # bug in maufbapi? tulir said: "ask meta"
        self.contacts.user_legacy_id = self.my_id
        await self.add_threads()
        self.mqtt.register_handlers()
        self.create_task(self.mqtt.listen(self.mqtt.seq_id))
        return f"Connected as '{self.me.name} <{self.me.email}>'"

    async def add_threads(self):
        thread_list = await self.api.fetch_thread_list(
            msg_count=0, thread_count=config.CHATS_TO_FETCH
        )
        self.mqtt.seq_id = int(thread_list.sync_sequence_id)  # type:ignore
        self.log.debug("SEQ ID: %s", self.mqtt.seq_id)
        self.log.debug("Thread list: %s", thread_list)
        self.log.debug("Thread list page info: %s", thread_list.page_info)
        contacts = []
        for t in thread_list.nodes:
            if t.is_group_thread:
                continue
            try:
                c = await self.contacts.by_thread(t)
            except XMPPError as e:
                self.log.warning(
                    "Something went wrong with this thread: %s", t, exc_info=e
                )
                continue
            contacts.append(c)
            await c.add_to_roster()
        for t in thread_list.nodes:
            if not t.is_group_thread:
                continue
            try:
                g = await self.bookmarks.by_thread(t)
            except XMPPError as e:
                self.log.warning(
                    "Something went wrong with this group thread: %s", t, exc_info=e
                )
                continue
            await g.add_to_bookmarks()
        for c in contacts:
            c.online()

    async def logout(self):
        pass

    async def on_text(
        self, chat: Recipient, text: str, *, reply_to_msg_id=None, **kwargs
    ) -> str:
        resp: mqtt_t.SendMessageResponse = await self.mqtt.send_message(
            target=chat.legacy_id,
            message=text,
            is_group=False,
            reply_to=reply_to_msg_id,
        )
        fut = self.ack_futures[
            resp.offline_threading_id
        ] = self.xmpp.loop.create_future()
        log.debug("Send message response: %s", resp)
        if not resp.success:
            raise XMPPError("internal-server-error", resp.error_message)
        fb_msg = await fut
        self.sent_messages[chat.legacy_id].add(fb_msg)
        return fb_msg.mid

    async def on_file(
        self, chat: Recipient, url: str, http_response, reply_to_msg_id=None, **_
    ):
        oti = self.mqtt.generate_offline_threading_id()
        fut = self.ack_futures[oti] = self.xmpp.loop.create_future()
        resp = await self.api.send_media(
            data=await http_response.read(),
            file_name=url.split("/")[-1],
            mimetype=http_response.content_type,
            offline_threading_id=oti,
            chat_id=chat.legacy_id,
            is_group=False,
            reply_to=reply_to_msg_id,
        )
        ack = await fut
        log.debug("Upload ack: %s", ack)
        return resp.media_id

    async def on_composing(self, c: Recipient, thread=None):
        await self.mqtt.set_typing(target=c.legacy_id)

    async def on_paused(self, c: Recipient, thread=None):
        await self.mqtt.set_typing(target=c.legacy_id, typing=False)

    async def on_displayed(self, c: Recipient, legacy_msg_id: str, thread=None):
        try:
            t = self.received_messages[c.legacy_id].by_mid[legacy_msg_id].timestamp_ms
        except KeyError:
            log.debug("Cannot find the timestamp of %s", legacy_msg_id)
        else:
            await self.mqtt.mark_read(target=c.legacy_id, read_to=t, is_group=False)

    async def on_react(
        self, c: Recipient, legacy_msg_id: str, emojis: list[str], thread=None
    ):
        # only one reaction per msg on facebook, but this is handled by slidge core
        if len(emojis) == 0:
            emoji = None
        else:
            emoji = emojis[-1]
        f = self.reaction_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        await self.api.react(legacy_msg_id, emoji)
        await f

    async def on_retract(self, c: Recipient, legacy_msg_id: str, thread=None):
        f = self.unsend_futures[legacy_msg_id] = self.xmpp.loop.create_future()
        await self.api.unsend(legacy_msg_id)
        await f

    async def on_search(self, form_values: dict[str, str]) -> SearchResult:
        results = await self.api.search(form_values["query"], entity_types=["user"])
        log.debug("Search results: %s", results)
        items = []
        for search_result in results.search_results.edges:
            result = search_result.node
            if isinstance(result, Participant):
                try:
                    contact = await self.contacts.by_legacy_id(int(result.id))
                except XMPPError:
                    items.append(
                        {
                            "name": result.name + " (problem)",
                            "jid": "",
                        }
                    )
                    continue
                if contact.is_friend:
                    await contact.add_to_roster()
                items.append(
                    {
                        "name": (
                            result.name + " (friend)"
                            if contact.is_friend
                            else " (not friend)"
                        ),
                        "jid": contact.jid.bare,
                    }
                )

        return SearchResult(
            fields=[
                FormField(var="name", label="Name"),
                FormField(var="jid", label="JID", type="jid-single"),
            ],
            items=items,
        )


log = logging.getLogger(__name__)



from datetime import datetime
from typing import TYPE_CHECKING, Optional

from maufbapi.types.graphql import Thread
from slidge import LegacyBookmarks, LegacyMUC, LegacyParticipant, MucType
from slidge.util.types import HoleBound
from slixmpp.exceptions import XMPPError

from .util import ChatterMixin

if TYPE_CHECKING:
    from .session import Session


class Bookmarks(LegacyBookmarks[int, "MUC"]):
    session: "Session"

    def __init__(self, session):
        super().__init__(session)
        self.threads = dict[int, Thread]()

    async def by_thread(self, thread: Thread):
        fb_id = int(thread.thread_key.thread_fbid)
        self.threads[fb_id] = thread
        return await self.by_legacy_id(fb_id)

    async def legacy_id_to_jid_local_part(self, i: int):
        return str(i)

    async def jid_local_part_to_legacy_id(self, local_part: str):
        try:
            return int(local_part)
        except ValueError:
            raise XMPPError(
                "item-not-found", f"Not a valid messenger chat id: {local_part}"
            )

    async def fill(self):
        pass


class MUC(LegacyMUC[int, str, "Participant", int]):
    session: "Session"

    type = MucType.GROUP
    REACTIONS_SINGLE_EMOJI = True
    _ALL_INFO_FILLED_ON_STARTUP = True

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.thread = None

    async def get_thread(self):
        thread = self.thread
        if not thread:
            thread = self.session.bookmarks.threads.pop(self.legacy_id, None)

        if not thread:
            try:
                threads = await self.session.api.fetch_thread_info(self.legacy_id)
            except Exception as e:
                raise XMPPError(
                    "internal-server-error", f"This group chat cannot be fetched: {e}"
                )
            if not threads:
                raise XMPPError("item-not-found")
            if not isinstance(threads, list) or len(threads) != 1:
                self.log.warning("Weird fetch_thread_info response: %s", threads)
                raise XMPPError(
                    "internal-server-error", f"Bad response from maufbapi: {threads}"
                )

            thread = threads[0]
        self.thread = thread
        return thread

    async def update_info(self, thread: Optional[Thread] = None):
        if not thread:
            thread = await self.get_thread()
        if thread.name:
            self.name = thread.name
        else:
            self.name = ", ".join(
                [
                    n.messaging_actor.name.split(" ")[0]
                    for n in thread.all_participants.nodes
                    if int(n.id) != self.session.contacts.user_legacy_id
                ]
            )
        if thread.image:
            if self.avatar != thread.image.uri:
                await self.set_avatar(thread.image.uri)
        else:
            self.avatar = None
        self.n_participants = len(thread.all_participants.nodes)

    async def fill_participants(self):
        t = await self.get_thread()
        self.log.debug("%s participants", len(t.all_participants.nodes))
        for p in t.all_participants.nodes:
            self.log.debug("participant: %s", p)
            if int(p.messaging_actor.id) == self.session.contacts.user_legacy_id:
                continue
            contact = await self.session.contacts.by_legacy_id(
                int(p.messaging_actor.id), p
            )
            part = await self.get_participant_by_contact(contact)
            self.log.debug("participant: %s", part)
            yield part

    async def backfill(
        self, after: Optional[HoleBound] = None, before: Optional[HoleBound] = None
    ):
        t = await self.get_thread()
        for m in t.messages.nodes:
            if not m.id:
                continue
            timestamp = datetime.fromtimestamp(m.timestamp / 1_000)
            if after is not None and timestamp < after.timestamp:
                continue
            if before is not None and timestamp > before.timestamp:
                continue
            part = await self.get_participant_by_legacy_id(
                int(m.message_sender.messaging_actor.id)
            )
            await part.send_fb_message(m)


class Participant(ChatterMixin, LegacyParticipant):
    @property
    def fb_id(self):
        return self.muc.legacy_id



import logging
from typing import TYPE_CHECKING, Optional

from maufbapi.types import mqtt as mqtt_t
from maufbapi.types.graphql import ParticipantNode, Thread
from maufbapi.types.graphql.responses import FriendshipStatus
from slidge import LegacyContact, LegacyRoster
from slixmpp.exceptions import XMPPError

from .util import ChatterMixin, is_group_thread

if TYPE_CHECKING:
    from .session import Session


class Contact(ChatterMixin, LegacyContact[int]):
    CORRECTION = False
    REACTIONS_SINGLE_EMOJI = True
    session: "Session"

    def __init__(
        self,
        session: "Session",
        legacy_id: int,
        jid_username: str,
        participant: Optional[ParticipantNode] = None,
    ):
        super().__init__(session, legacy_id, jid_username)
        if participant is not None:
            self.populate_from_participant(participant)

    @property
    def fb_id(self):
        return self.legacy_id

    def populate_from_participant(
        self, participant: ParticipantNode, update_avatar=True
    ):
        if self.legacy_id != int(participant.messaging_actor.id):
            raise XMPPError(
                "bad-request",
                (
                    f"Legacy ID {self.legacy_id} does not match participant"
                    f" {participant.messaging_actor.id}"
                ),
            )
        self.is_friend = (
            participant.messaging_actor.friendship_status
            == FriendshipStatus.ARE_FRIENDS
        )
        self.name = participant.messaging_actor.name
        if self.avatar is None or update_avatar:
            pic = participant.messaging_actor.profile_pic_large
            self.avatar = pic.uri

    async def get_thread(self, **kwargs):
        threads = await self.session.api.fetch_thread_info(self.legacy_id, **kwargs)
        if len(threads) != 1:
            self.log.debug("Could not determine my profile! %s", threads)
            raise XMPPError(
                "internal-server-error",
                f"The messenger API returned {len(threads)} threads for this user.",
            )
        return threads[0]

    async def update_info(self, refresh=False):
        if self.name and not refresh:
            return
        t = await self.get_thread(msg_count=0)

        participant = self.session.contacts.get_friend_participant(
            t.all_participants.nodes
        )
        self.populate_from_participant(participant)


class Roster(LegacyRoster[int, Contact]):
    session: "Session"

    async def by_thread_key(self, t: mqtt_t.ThreadKey):
        if is_group_thread(t):
            raise ValueError("Thread seems to be a group thread")
        c = await self.by_legacy_id(t.other_user_id)
        await c.add_to_roster()
        return c

    async def by_thread(self, t: Thread):
        if t.is_group_thread:
            raise XMPPError(
                "bad-request", f"Legacy ID {t.id} is a group chat, not a contact"
            )

        participant = self.get_friend_participant(t.all_participants.nodes)
        fb_id = int(participant.messaging_actor.id)
        contact = await self.by_legacy_id(fb_id, participant)
        return contact

    def get_friend_participant(self, nodes: list[ParticipantNode]) -> ParticipantNode:
        if len(nodes) != 2:
            raise XMPPError(
                "internal-server-error",
                (
                    "This facebook thread has more than two participants. This is a"
                    " slidge bug."
                ),
            )

        for participant in nodes:
            if int(participant.id) != self.session.my_id:
                return participant
        else:
            raise XMPPError(
                "internal-server-error", "Couldn't find friend in thread participants"
            )


log = logging.getLogger(__name__)



import maufbapi.http.api
from maufbapi.types.graphql import OwnInfo
from slidge.util.util import get_version  # noqa: F401

from . import config, contact, gateway, group, session

# workaround until https://github.com/mautrix/facebook/pull/318 is merged and
# released
maufbapi.http.api.OwnInfo = OwnInfo

__all__ = "config", "contact", "gateway", "group", "session"

__version__ = "v0.2.0"



from typing import TYPE_CHECKING, Optional

from maufbapi import AndroidMQTT as AndroidMQTTOriginal
from maufbapi.types import mqtt as mqtt_t
from slidge.util.types import MessageReference

from .util import FacebookMessage, is_group_thread

REQUEST_TIMEOUT = 60

if TYPE_CHECKING:
    from .session import Session


class AndroidMQTT(AndroidMQTTOriginal):
    def __init__(
        self,
        session: "Session",
        state,
        loop=None,
        log=None,
        connect_token_hash=None,
        proxy_handler=None,
    ):
        # overrides the default init to enable presences
        self.session = session
        super().__init__(state, loop, log, connect_token_hash, proxy_handler)

    def register_handlers(self):
        self.seq_id_update_callback = lambda i: setattr(  # type:ignore
            self, "seq_id", i
        )
        self.add_event_handler(mqtt_t.Message, self.on_fb_message)
        self.add_event_handler(mqtt_t.ExtendedMessage, self.on_fb_extended_message)
        self.add_event_handler(mqtt_t.ReadReceipt, self.on_fb_message_read)
        self.add_event_handler(mqtt_t.TypingNotification, self.on_fb_typing)
        self.add_event_handler(mqtt_t.OwnReadReceipt, self.on_fb_user_read)
        self.add_event_handler(mqtt_t.Reaction, self.on_fb_reaction)
        self.add_event_handler(mqtt_t.UnsendMessage, self.on_fb_unsend)

        self.add_event_handler(mqtt_t.AvatarChange, self.on_fb_event)
        self.add_event_handler(mqtt_t.AddMember, self.on_fb_event)
        self.add_event_handler(mqtt_t.RemoveMember, self.on_fb_event)

        self.add_event_handler(mqtt_t.NameChange, self.on_fb_event)
        self.add_event_handler(mqtt_t.ThreadChange, self.on_fb_event)
        self.add_event_handler(mqtt_t.MessageSyncError, self.on_fb_event)
        self.add_event_handler(mqtt_t.ForcedFetch, self.on_fb_event)

    async def __get_chatter(
        self, thread_key: mqtt_t.ThreadKey, fb_id: Optional[int] = None
    ):
        if is_group_thread(thread_key):
            self.log.debug("MUC")
            muc = await self.session.bookmarks.by_legacy_id(thread_key.thread_fbid)
            self.log.debug("MUC %s", muc)
            if fb_id:
                self.log.debug("fbid: %s", fb_id)
                chatter = await muc.get_participant_by_legacy_id(fb_id)
            else:
                self.log.debug("no fbid: %s", fb_id)
                chatter = await muc.get_user_participant()
            self.log.debug("chatter: %s", chatter)
        else:
            chatter = await self.session.contacts.by_thread_key(thread_key)
        return chatter

    async def on_fb_extended_message(self, evt: mqtt_t.ExtendedMessage):
        self.log.debug("Extended message")
        kwargs = {}
        msg = evt.message

        if reply_to_fb_msg := evt.reply_to_message:
            self.log.debug("Reply-to")
            author_fb_id = reply_to_fb_msg.metadata.sender
            if author_fb_id == self.session.my_id:
                author = "user"  # type: ignore
            else:
                author = await self.session.contacts.by_legacy_id(author_fb_id)  # type: ignore
            kwargs["reply_to"] = MessageReference(
                legacy_id=reply_to_fb_msg.metadata.id,
                body=reply_to_fb_msg.text,
                author=author,  # type: ignore
            )
        await self.on_fb_message(msg, **kwargs)

    async def on_fb_message(self, msg: mqtt_t.Message, **kwargs):
        meta = msg.metadata
        chatter = await self.__get_chatter(meta.thread, meta.sender)
        fb_msg = FacebookMessage(mid=meta.id, timestamp_ms=meta.timestamp)

        if meta.sender == self.session.my_id:
            try:
                fut = self.session.ack_futures.pop(meta.offline_threading_id)
            except KeyError:
                self.log.debug("Received carbon %s - %s", meta.id, msg.text)
                kwargs["carbon"] = True
                self.log.debug("Sent carbon")
                self.session.sent_messages[chatter.fb_id].add(fb_msg)
            else:
                self.log.debug("Received echo of %s", meta.offline_threading_id)
                fut.set_result(fb_msg)
                return
        else:
            self.session.received_messages[chatter.fb_id].add(fb_msg)

        await chatter.send_fb_message(msg, **kwargs)

    async def on_fb_message_read(self, receipt: mqtt_t.ReadReceipt):
        self.log.debug("Facebook read: %s", receipt)
        chatter = await self.__get_chatter(receipt.thread, receipt.user_id)
        try:
            mid = (
                self.session.sent_messages[chatter.fb_id].pop_up_to(receipt.read_to).mid
            )
        except KeyError:
            self.log.debug("Cannot find MID of %s", receipt.read_to)
        else:
            chatter.displayed(mid)

    async def on_fb_typing(self, notification: mqtt_t.TypingNotification):
        self.log.debug("Facebook typing: %s", notification)
        c = await self.session.contacts.by_legacy_id(notification.user_id)
        if notification.typing_status:
            c.composing()
        else:
            c.paused()

    async def on_fb_user_read(self, receipt: mqtt_t.OwnReadReceipt):
        self.log.debug("Facebook own read: %s", receipt)
        when = receipt.read_to
        for thread in receipt.threads:
            c = await self.__get_chatter(thread)
            try:
                mid = self.session.received_messages[c.fb_id].pop_up_to(when).mid
            except KeyError:
                self.log.debug("Cannot find mid of %s", when)
                continue
            c.displayed(mid, carbon=True)

    async def on_fb_reaction(self, reaction: mqtt_t.Reaction):
        self.log.debug("Reaction: %s", reaction)
        chatter = await self.__get_chatter(reaction.thread, reaction.reaction_sender_id)
        mid = reaction.message_id
        if reaction.reaction_sender_id == self.session.my_id:
            try:
                f = self.session.reaction_futures.pop(mid)
            except KeyError:
                chatter.react(mid, reaction.reaction or "", carbon=True)
            else:
                f.set_result(None)
        else:
            chatter.react(reaction.message_id, reaction.reaction or "")

    async def on_fb_unsend(self, unsend: mqtt_t.UnsendMessage):
        self.log.debug("Unsend: %s", unsend)
        chatter = await self.__get_chatter(unsend.thread, unsend.user_id)
        mid = unsend.message_id
        if unsend.user_id == self.session.my_id:
            try:
                f = self.session.unsend_futures.pop(mid)
            except KeyError:
                chatter.retract(mid, carbon=True)
            else:
                f.set_result(None)
        else:
            chatter.retract(unsend.message_id)

    async def on_fb_event(self, evt):
        self.log.debug("Facebook event: %s", evt)



import random
from typing import Optional

import maufbapi.http
from maufbapi import AndroidAPI, AndroidState
from mautrix.util.proxy import ProxyHandler
from slidge import BaseGateway, FormField, GatewayUser
from slidge.command.register import RegistrationType, TwoFactorNotRequired
from slixmpp import JID
from slixmpp.exceptions import XMPPError

from .util import save_state


class Gateway(BaseGateway):
    REGISTRATION_INSTRUCTIONS = "Enter facebook credentials"
    REGISTRATION_FIELDS = [
        FormField(var="email", label="Email", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]
    REGISTRATION_MULTISTEP = True
    REGISTRATION_TYPE = RegistrationType.TWO_FACTOR_CODE

    ROSTER_GROUP = "Facebook"

    COMPONENT_NAME = "Facebook (slidge)"
    COMPONENT_TYPE = "facebook"
    COMPONENT_AVATAR = (
        "https://upload.wikimedia.org/wikipedia/commons/thumb/b/be/"
        "Facebook_Messenger_logo_2020.svg/640px-Facebook_Messenger_logo_2020.svg.png"
    )

    SEARCH_TITLE = "Search in your facebook friends"
    SEARCH_INSTRUCTIONS = (
        "Enter something that can be used to search for one of your friends, eg, a"
        " first name"
    )
    SEARCH_FIELDS = [FormField(var="query", label="Search term(s)", required=True)]

    GROUPS = True

    LEGACY_CONTACT_ID_TYPE = int

    def __init__(self):
        super().__init__()
        self._pending_reg = dict[str, AndroidAPI]()

    async def validate(
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ):
        s = AndroidState()
        x = ProxyHandler(None)
        api = AndroidAPI(state=s, proxy_handler=x)
        s.generate(random.randbytes(30))  # type: ignore
        await api.mobile_config_sessionless()
        try:
            await api.login(
                email=registration_form["email"], password=registration_form["password"]
            )
        except maufbapi.http.errors.TwoFactorRequired:
            self._pending_reg[user_jid.bare] = api
        except maufbapi.http.errors.OAuthException as e:
            raise XMPPError("not-authorized", text=str(e))
        else:
            save_state(user_jid.bare, api.state)
            raise TwoFactorNotRequired

    async def validate_two_factor_code(self, user: GatewayUser, code):
        api = self._pending_reg.pop(user.jid.bare)
        try:
            await api.login_2fa(email=user.legacy_module_data["email"], code=code)
        except maufbapi.http.errors as e:
            raise XMPPError("not-authorized", text=str(e))
        save_state(user.jid.bare, api.state)



from slidge.util.test import SlidgeTest

import messlidger


class TestMesslidger(SlidgeTest):
    def test_base(self):
        self.recv("<presence />")
        reply = self.next_sent()
        assert reply["type"] == "error"



import unittest.mock

from messlidger.util import FacebookMessage, Messages, ChatterMixin


def test_find_closest_timestamp():
    sent = Messages()
    sent.add(FacebookMessage("a", 2))
    sent.add(FacebookMessage("b", 5))
    sent.add(FacebookMessage("c", 10))
    sent.add(FacebookMessage("d", 23))
    sent.add(FacebookMessage("e", 45))
    m = sent.pop_up_to(15)
    assert m.timestamp_ms == 10
    assert m.mid == "c"
    assert len(sent) == 2
    assert "d" in sent.by_mid
    assert "e" in sent.by_mid

    m = sent.pop_up_to(500)
    assert m.timestamp_ms == 45
    assert m.mid == "e"
    assert len(sent) == 0

    try:
        sent.pop_up_to(5)
    except KeyError as e:
        assert e.args == (5,)

    assert len(sent.by_timestamp_ms) == 0


def test_extensible_media():
    text, urls = ChatterMixin.get_extensible_media(
        unittest.mock.MagicMock,
        """{
      "ZXh0ZW5zaWJsZV9tZXNzYWdlX2F0dGFjaG1lbnQ6ZWUubWlkLiRjQUFBQUFBQlJjclNTc3ZzUUVXTWdiWExjZHU4Sg": {
        "__typename": "ExtensibleMessageAttachment",
        "strong_id__": "ZXh0ZW5zaWJsZV9tZXNzYWdlX2F0dGFjaG1lbnQ6ZWUubWlkLiRjQUFBQUFBQlJjclNTc3ZzUUVXTWdiWExjZHU4Sg",
        "id": "ZXh0ZW5zaWJsZV9tZXNzYWdlX2F0dGFjaG1lbnQ6ZWUubWlkLiRjQUFBQUFBQlJjclNTc3ZzUUVXTWdiWExjZHU4Sg",
        "is_forwardable": true,
        "story_attachment": {
          "url": "https:\\/\\/www.facebook.com\\/gelati.arlequin",
          "title": "Arlequin Gelati",
          "target": {
            "__typename": "User",
            "strong_id__": "100063607510573",
            "id": "100063607510573",
            "friendship_status": "CANNOT_REQUEST"
          },
          "media": {
            "__typename": "GenericAttachmentMedia",
            "imageLarge": {
              "uri": "https:\\/\\/scontent.xx.fbcdn.net\\/v\\/t39.30808-1\\/277748953_396876902442557_5384670406191252860_n.jpg?stp=dst-jpg_p100x100&_nc_cat=102&ccb=1-7&_nc_sid=3a5e4d&_nc_ohc=cMZjLhTqWCoAX8y2EDh&_nc_ad=z-m&_nc_cid=0&_nc_ht=scontent.xx&oh=00_AfDSMCokBsH3Q377ydXCx5oj4Mz3DdY5liEGlvvQxPA9Ig&oe=6586A89A"
            },
            "image": {
              "uri": "https:\\/\\/scontent.xx.fbcdn.net\\/v\\/t39.30808-1\\/277748953_396876902442557_5384670406191252860_n.jpg?stp=dst-jpg_p100x100&_nc_cat=102&ccb=1-7&_nc_sid=3a5e4d&_nc_ohc=cMZjLhTqWCoAX8y2EDh&_nc_ad=z-m&_nc_cid=0&_nc_ht=scontent.xx&oh=00_AfDSMCokBsH3Q377ydXCx5oj4Mz3DdY5liEGlvvQxPA9Ig&oe=6586A89A"
            },
            "imageNatural": {
              "uri": "https:\\/\\/scontent.xx.fbcdn.net\\/v\\/t39.30808-1\\/277748953_396876902442557_5384670406191252860_n.jpg?stp=dst-jpg_p100x100&_nc_cat=102&ccb=1-7&_nc_sid=3a5e4d&_nc_ohc=cMZjLhTqWCoAX8y2EDh&_nc_ad=z-m&_nc_cid=0&_nc_ht=scontent.xx&oh=00_AfDSMCokBsH3Q377ydXCx5oj4Mz3DdY5liEGlvvQxPA9Ig&oe=6586A89A"
            },
            "imageFullScreen": {
              "uri": "https:\\/\\/scontent.xx.fbcdn.net\\/v\\/t39.30808-1\\/277748953_396876902442557_5384670406191252860_n.jpg?stp=dst-jpg_p100x100&_nc_cat=102&ccb=1-7&_nc_sid=3a5e4d&_nc_ohc=cMZjLhTqWCoAX8y2EDh&_nc_ad=z-m&_nc_cid=0&_nc_ht=scontent.xx&oh=00_AfDSMCokBsH3Q377ydXCx5oj4Mz3DdY5liEGlvvQxPA9Ig&oe=6586A89A"
            },
            "id": ""
          },
          "style_list": [
            "share",
            "fallback"
          ],
          "title_with_entities": {
            "text": "Arlequin Gelati"
          },
          "description": {
            "text": "Glacier artisanal gastronomique"
          },
          "action_links": [
            {
              "__typename": "LinkOpenActionLink",
              "title": "No button"
            }
          ],
          "style_infos": [
            {
              "__typename": "FeedStandardAttachmentStyleInfo"
            },
            {
              "__typename": "FutureOfFeedAttachmentStyleInfo"
            }
          ],
          "deduplication_key": "e9c579b2cb99fcb2f9ddcec4cf4456a6"
        },
        "genie_attachment": {
          "genie_message": {
            "__typename": "User",
            "strong_id__": "100063607510573"
          }
        }
      }
    }
    """,
    )
    assert (
        text
        == "\nhttps://www.facebook.com/gelati.arlequin - Arlequin Gelati - Glacier artisanal gastronomique"
    )
    assert (
        "https://scontent.xx.fbcdn.net/v/t39.30808-1/277748953_396876902442557_5384670406191252860_n.jpg?stp=dst-jpg_p100x100&_nc_cat=102&ccb=1-7&_nc_sid=3a5e4d&_nc_ohc=cMZjLhTqWCoAX8y2EDh&_nc_ad=z-m&_nc_cid=0&_nc_ht=scontent.xx&oh=00_AfDSMCokBsH3Q377ydXCx5oj4Mz3DdY5liEGlvvQxPA9Ig&oe=6586A89A"
        in urls
    )

    text, urls = ChatterMixin.get_extensible_media(
        unittest.mock.MagicMock,
        """
        {
  "ZXh0ZW5zaWJsZV9tZXNzYWdlX2F0dGFjaG1lbnQ6ZWUubWlkLiRjQUFBN1RSMlYwcWFRSnA1ZC1XSjlWa0duaE1QSw": {
    "typename": "ExtensibleMessageAttachment",
    "strong_id": "ZXh0ZW5zaWJsZV9tZXNzYWdlX2F0dGFjaG1lbnQ6ZWUubWlkLiRjQUFBN1RSMlYwcWFRSnA1ZC1XSjlWa0duaE1QSw",
    "id": "ZXh0ZW5zaWJsZV9tZXNzYWdlX2F0dGFjaG1lbnQ6ZWUubWlkLiRjQUFBN1RSMlYwcWFRSnA1ZC1XSjlWa0duaE1QSw",
    "is_forwardable": true,
    "story_attachment": {
      "url": "fbrpc://facebook/nativethirdparty?app_id=256002347743983&app_name=Facebook+Messenger+for+Android&fallback_url0=https%3A%2F%2Fapps.facebook.com%2Ffbmessenger_android%2F&market_uri=market%3A%2F%2Fdetails%3Fid%3Dcom.facebook.orca%26referrer%3Dutm_source%253Dapps.facebook.com%2526utm_campaign%253Dfb4a%2526utm_content%253D%25257B%252522app%252522%25253A256002347743983%25252C%252522t%252522%25253A1692038403%25252C%252522source%252522%25253Anull%25257D%26app_id%3D256002347743983%26is_vt_odir_eligible%3D0&package_name=com.facebook.orca&tap_behavior=app_fallback_web&target_url=https%3A%2F%2Fwww.amazon.com%2FBCW-1-TLCH-100-Topload-Card-Holder%2Fdp%2FB074VH2VWV%2Fref%3Dasc_df_B074VH2VWV%2F%3Ftag%3Dhyprod-20%26linkCode%3Ddf0%26hvadid%3D241968535606%26hvpos%26hvnetw%3Dg%26hvrand%3D17835882389631259700%26hvpone%26hvptwo%26hvqmt%26hvdev%3Dm%26hvdvcmdl%26hvlocint%26hvlocphy%3D9022116%26hvtargid%3Dpla-600746527901%26psc%3D1&extra_applink_key=al_applink_data&referer_data_key=extras&al_applink_data=%7B%22target_url%22%3A%22https%3A%5C%2F%5C%2Fwww.amazon.com%5C%2FBCW-1-TLCH-100-Topload-Card-Holder%5C%2Fdp%5C%2FB074VH2VWV%5C%2Fref%3Dasc_df_B074VH2VWV%5C%2F%3Ftag%3Dhyprod-20%26linkCode%3Ddf0%26hvadid%3D241968535606%26hvpos%26hvnetw%3Dg%26hvrand%3D17835882389631259700%26hvpone%26hvptwo%26hvqmt%26hvdev%3Dm%26hvdvcmdl%26hvlocint%26hvlocphy%3D9022116%26hvtargid%3Dpla-600746527901%26psc%3D1%22%2C%22extras%22%3A%7B%22fb_app_id%22%3A256002347743983%7D%2C%22referer_app_link%22%3A%7B%22url%22%3A%22fb-messenger%3A%5C%2F%5C%2F%5C%2F%22%2C%22app_name%22%3A%22Messenger%22%2C%22package%22%3A%22com.facebook.orca%22%7D%7D&appsite_data=%7B%22android%22%3A%5B%7B%22is_app_link%22%3Afalse%2C%22app_name%22%3A%22Facebook+Messenger+for+Android%22%2C%22fallback_url%22%3A%22https%3A%5C%2F%5C%2Fapps.facebook.com%5C%2Ffbmessenger_android%5C%2F%22%2C%22market_uri%22%3A%22market%3A%5C%2F%5C%2Fdetails%3Fid%3Dcom.facebook.orca%26referrer%3Dutm_source%5Cu00253Dapps.facebook.com%5Cu002526utm_campaign%5Cu00253Dfb4a%5Cu002526utm_content%5Cu00253D%5Cu0025257B%5Cu00252522app%5Cu00252522%5Cu0025253A256002347743983%5Cu0025252C%5Cu00252522t%5Cu00252522%5Cu0025253A1692038403%5Cu0025252C%5Cu00252522source%5Cu00252522%5Cu0025253Anull%5Cu0025257D%26app_id%3D256002347743983%26is_vt_odir_eligible%3D0%22%2C%22package%22%3A%22com.facebook.orca%22%7D%5D%7D&has_app_link=0",
      "title": "BCW 1-TLCH-100 3X4 Topload Card Holder - Standard",
      "target": {
        "typename": "ExternalUrl",
        "strong_id": "-3483010280417037662",
        "cache_id": "-3483010280417037662"
      },
      "media": {
        "__typename": "GenericAttachmentMedia",
        "imageLarge": {
          "uri": "https://external.xx.fbcdn.net/emg1/v/t13/1740253623741233594?url=https%3A%2F%2Fm.media-amazon.com%2Fimages%2FI%2F51Bdv%252BAUTpL._SR600%252c315_PIWhiteStrip%252cBottomLeft%252c0%252c35_PIStarRatingFOURANDHALF%252cBottomLeft%252c360%252c-6_SR600%252c315_ZA1%25252C081%252c445%252c290%252c400%252c400%252cAmazonEmberBold%252c12%252c4%252c0%252c0%252c5_SCLZZZZZZZ_FMpng_BG255%252c255%252c255.jpg&fb_obo=1&utld=media-amazon.com&stp=c0.5000x0.5000f_dst-emg0_p600x300_q75&ccb=13-1&oh=06_AbHdzVVTmuPsTJnfzPax8Jh8VaeGbFbLUHGZo9IzqUOStQ&oe=64DC008F&_nc_sid=cab974",
          "width": 600,
          "height": 300
        },
        "image": {
          "uri": "https://external.xx.fbcdn.net/emg1/v/t13/1740253623741233594?url=https%3A%2F%2Fm.media-amazon.com%2Fimages%2FI%2F51Bdv%252BAUTpL._SR600%252c315_PIWhiteStrip%252cBottomLeft%252c0%252c35_PIStarRatingFOURANDHALF%252cBottomLeft%252c360%252c-6_SR600%252c315_ZA1%25252C081%252c445%252c290%252c400%252c400%252cAmazonEmberBold%252c12%252c4%252c0%252c0%252c5_SCLZZZZZZZ_FMpng_BG255%252c255%252c255.jpg&fb_obo=1&utld=media-amazon.com&stp=c0.5000x0.5000f_dst-emg0_p600x300_q75&ccb=13-1&oh=06_AbHdzVVTmuPsTJnfzPax8Jh8VaeGbFbLUHGZo9IzqUOStQ&oe=64DC008F&_nc_sid=cab974",
          "width": 600,
          "height": 300
        },
        "imageNatural": {
          "uri": "https://external.xx.fbcdn.net/emg1/v/t13/1740253623741233594?url=https%3A%2F%2Fm.media-amazon.com%2Fimages%2FI%2F51Bdv%252BAUTpL._SR600%252c315_PIWhiteStrip%252cBottomLeft%252c0%252c35_PIStarRatingFOURANDHALF%252cBottomLeft%252c360%252c-6_SR600%252c315_ZA1%25252C081%252c445%252c290%252c400%252c400%252cAmazonEmberBold%252c12%252c4%252c0%252c0%252c5_SCLZZZZZZZ_FMpng_BG255%252c255%252c255.jpg&fb_obo=1&utld=media-amazon.com&stp=c0.5000x0.5000f_dst-emg0_p315x315_q75&ccb=13-1&oh=06_AbHvbZ_aX5MJ_5KwbBv5bMsFjUsNarr3E79K3WUUJ42yOA&oe=64DC008F&_nc_sid=cab974",
          "width": 315,
          "height": 315
        },
        "imageFullScreen": {
          "uri": "https://external.xx.fbcdn.net/emg1/v/t13/1740253623741233594?url=https%3A%2F%2Fm.media-amazon.com%2Fimages%2FI%2F51Bdv%252BAUTpL._SR600%252c315_PIWhiteStrip%252cBottomLeft%252c0%252c35_PIStarRatingFOURANDHALF%252cBottomLeft%252c360%252c-6_SR600%252c315_ZA1%25252C081%252c445%252c290%252c400%252c400%252cAmazonEmberBold%252c12%252c4%252c0%252c0%252c5_SCLZZZZZZZ_FMpng_BG255%252c255%252c255.jpg&fb_obo=1&utld=media-amazon.com&stp=c0.5000x0.5000f_dst-emg0_p315x315_q75&ccb=13-1&oh=06_AbHvbZ_aX5MJ_5KwbBv5bMsFjUsNarr3E79K3WUUJ42yOA&oe=64DC008F&_nc_sid=cab974",
          "width": 315,
          "height": 315
        }
      },
      "style_list": [
        "share",
        "fallback"
      ],
      "title_with_entities": {
        "text": "BCW 1-TLCH-100 3X4 Topload Card Holder - Standard"
      },
      "description": {
        "text": "BCW standard 3x4 top loading card holders are made of high-quality, rigid PVC. These top loading holders are some of the most popular rigid individual card holders in the collectible card industry. Us"
      },
      "source": {
        "text": "amazon.com"
      },
      "attachment_properties": [
        {
          "key": "width",
          "attachment_property_type": "double",
          "value": {
            "text": "600"
          }
        },
        {
          "key": "height",
          "attachment_property_type": "double",
          "value": {
            "text": "315"
          }
        }
      ],
      "style_infos": [
        {
          "__typename": "FeedStandardAttachmentStyleInfo"
        },
        {
          "__typename": "ExternalShareAttachmentStyleInfo"
        },
        {
          "__typename": "FutureOfFeedAttachmentStyleInfo"
        }
      ],
      "deduplication_key": "ee.mid.$cAAA7TR2V0qaQJp5d-WJ9VkGnhMPK"
    },
    "genie_attachment": {
      "genie_message": {
        "typename": "ExternalUrl",
        "strong_id": "-3483010280417037662"
      }
    }
  }
}
        """,
    )
    assert "BCW 1-TLCH-100 3X4 Topload Card Holder - Standard" in text
    assert (
        "fbrpc://facebook/nativethirdparty?app_id=256002347743983&app_name=Facebook+Messenger+for+Android&fallback_url0=https%3A%2F%2Fapps.facebook.com%2Ffbmessenger_android%2F&market_uri=market%3A%2F%2Fdetails%3Fid%3Dcom.facebook.orca%26referrer%3Dutm_source%253Dapps.facebook.com%2526utm_campaign%253Dfb4a%2526utm_content%253D%25257B%252522app%252522%25253A256002347743983%25252C%252522t%252522%25253A1692038403%25252C%252522source%252522%25253Anull%25257D%26app_id%3D256002347743983%26is_vt_odir_eligible%3D0&package_name=com.facebook.orca&tap_behavior=app_fallback_web&target_url=https%3A%2F%2Fwww.amazon.com%2FBCW-1-TLCH-100-Topload-Card-Holder%2Fdp%2FB074VH2VWV%2Fref%3Dasc_df_B074VH2VWV%2F%3Ftag%3Dhyprod-20%26linkCode%3Ddf0%26hvadid%3D241968535606%26hvpos%26hvnetw%3Dg%26hvrand%3D17835882389631259700%26hvpone%26hvptwo%26hvqmt%26hvdev%3Dm%26hvdvcmdl%26hvlocint%26hvlocphy%3D9022116%26hvtargid%3Dpla-600746527901%26psc%3D1&extra_applink_key=al_applink_data&referer_data_key=extras&al_applink_data=%7B%22target_url%22%3A%22https%3A%5C%2F%5C%2Fwww.amazon.com%5C%2FBCW-1-TLCH-100-Topload-Card-Holder%5C%2Fdp%5C%2FB074VH2VWV%5C%2Fref%3Dasc_df_B074VH2VWV%5C%2F%3Ftag%3Dhyprod-20%26linkCode%3Ddf0%26hvadid%3D241968535606%26hvpos%26hvnetw%3Dg%26hvrand%3D17835882389631259700%26hvpone%26hvptwo%26hvqmt%26hvdev%3Dm%26hvdvcmdl%26hvlocint%26hvlocphy%3D9022116%26hvtargid%3Dpla-600746527901%26psc%3D1%22%2C%22extras%22%3A%7B%22fb_app_id%22%3A256002347743983%7D%2C%22referer_app_link%22%3A%7B%22url%22%3A%22fb-messenger%3A%5C%2F%5C%2F%5C%2F%22%2C%22app_name%22%3A%22Messenger%22%2C%22package%22%3A%22com.facebook.orca%22%7D%7D&appsite_data=%7B%22android%22%3A%5B%7B%22is_app_link%22%3Afalse%2C%22app_name%22%3A%22Facebook+Messenger+for+Android%22%2C%22fallback_url%22%3A%22https%3A%5C%2F%5C%2Fapps.facebook.com%5C%2Ffbmessenger_android%5C%2F%22%2C%22market_uri%22%3A%22market%3A%5C%2F%5C%2Fdetails%3Fid%3Dcom.facebook.orca%26referrer%3Dutm_source%5Cu00253Dapps.facebook.com%5Cu002526utm_campaign%5Cu00253Dfb4a%5Cu002526utm_content%5Cu00253D%5Cu0025257B%5Cu00252522app%5Cu00252522%5Cu0025253A256002347743983%5Cu0025252C%5Cu00252522t%5Cu00252522%5Cu0025253A1692038403%5Cu0025252C%5Cu00252522source%5Cu00252522%5Cu0025253Anull%5Cu0025257D%26app_id%3D256002347743983%26is_vt_odir_eligible%3D0%22%2C%22package%22%3A%22com.facebook.orca%22%7D%5D%7D&has_app_link=0"
        in text
    )
    assert (
        "https://external.xx.fbcdn.net/emg1/v/t13/1740253623741233594?url=https%3A%2F%2Fm.media-amazon.com%2Fimages%2FI%2F51Bdv%252BAUTpL._SR600%252c315_PIWhiteStrip%252cBottomLeft%252c0%252c35_PIStarRatingFOURANDHALF%252cBottomLeft%252c360%252c-6_SR600%252c315_ZA1%25252C081%252c445%252c290%252c400%252c400%252cAmazonEmberBold%252c12%252c4%252c0%252c0%252c5_SCLZZZZZZZ_FMpng_BG255%252c255%252c255.jpg&fb_obo=1&utld=media-amazon.com&stp=c0.5000x0.5000f_dst-emg0_p600x300_q75&ccb=13-1&oh=06_AbHdzVVTmuPsTJnfzPax8Jh8VaeGbFbLUHGZo9IzqUOStQ&oe=64DC008F&_nc_sid=cab974"
        in urls
    )



from datetime import datetime
from pathlib import Path

project = "messlidger"
copyright = f"{datetime.today().year}, the {project} contributors"
author = "the slidge contributors"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc.typehints",
    "autoapi.extension",
    "slidge_sphinx_extensions.doap",
    "slidge_sphinx_extensions.config_obj",
    "sphinx_mdinclude",
]

autodoc_typehints = "description"

# Include __init__ docstrings
autoclass_content = "both"
autoapi_python_class_content = "both"

autoapi_type = "python"
autoapi_dirs = [f"../../{project}"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "slixmpp": ("https://slixmpp.readthedocs.io/en/latest/", None),
    "slidge": ("https://slidge.im/docs/slidge/main/", None),
}

extlinks = {"xep": ("https://xmpp.org/extensions/xep-%s.html", "XEP-%s")}

html_theme = "furo"
html_theme_options = {
    "source_edit_link": f"https://codeberg.org/slidge/{project}/_edit/main/docs/source/{{filename}}",
    "source_view_link": f"https://codeberg.org/slidge/{project}/src/branch/main/docs/source/{{filename}}",
    "footer_icons": [
        {
            "name": "Codeberg",
            "url": f"https://codeberg.org/slidge/{project}",
            "html": Path("codeberg.svg").read_text(),
        },
    ],
}



CHATS_TO_FETCH = 20
CHATS_TO_FETCH__DOC = (
    "The number of most recent chats to fetch on startup. "
    "Getting all chats might hit rate limiting and possibly account lock. "
    "Please report if you try with high values and don't hit any problem!"
)



import json
import shelve
from collections import OrderedDict
from dataclasses import dataclass
from typing import TYPE_CHECKING

from maufbapi import AndroidState
from maufbapi.types import mqtt as mqtt_t
from slidge import global_config
from slidge.core.mixins.attachment import AttachmentMixin
from slidge.util.types import LegacyAttachment

if TYPE_CHECKING:
    from .session import Session


def get_shelf_path(user_bare_jid):
    return str(global_config.HOME_DIR / user_bare_jid)


def save_state(user_bare_jid: str, state: AndroidState):
    shelf_path = get_shelf_path(user_bare_jid)
    with shelve.open(shelf_path) as shelf:
        shelf["state"] = state


@dataclass
class FacebookMessage:
    mid: str
    timestamp_ms: int


class Messages:
    def __init__(self):
        self.by_mid: OrderedDict[str, FacebookMessage] = OrderedDict()
        self.by_timestamp_ms: OrderedDict[int, FacebookMessage] = OrderedDict()

    def __len__(self):
        return len(self.by_mid)

    def add(self, m: FacebookMessage):
        self.by_mid[m.mid] = m
        self.by_timestamp_ms[m.timestamp_ms] = m

    def pop_up_to(self, approx_t: int) -> FacebookMessage:
        i = 0
        for i, t in enumerate(self.by_timestamp_ms.keys()):
            if t > approx_t:
                i -= 1
                break
        for j, t in enumerate(list(self.by_timestamp_ms.keys())):
            msg = self.by_timestamp_ms.pop(t)
            self.by_mid.pop(msg.mid)
            if j == i:
                return msg
        else:
            raise KeyError(approx_t)


def is_group_thread(t: mqtt_t.ThreadKey):
    return t.other_user_id is None and t.thread_fbid is not None


class ChatterMixin(AttachmentMixin):
    session: "Session"

    async def send_fb_message(self, msg: mqtt_t.Message, **kwargs):
        kwargs["legacy_msg_id"] = msg.metadata.id

        sticker = msg.sticker
        if sticker is not None:
            return await self.send_fb_sticker(sticker, **kwargs)

        text = msg.text or ""
        atts = []
        for a in msg.attachments:
            url = await self.get_attachment_url(a, msg.metadata.thread, msg.metadata.id)
            if url:
                atts.append(
                    LegacyAttachment(
                        name=a.file_name,
                        content_type=a.mime_type,
                        url=url,
                        legacy_file_id=a.media_id_str,
                    )
                )
            if not a.extensible_media:
                continue
            try:
                subtext, urls = self.get_extensible_media(a.extensible_media)
            except ValueError:
                self.send_text(f"/me sent something slidge does not understand: {a}")
                continue
            text += subtext
            atts.extend(
                LegacyAttachment(url=url, content_type="image/jpeg") for url in urls
            )

        await self.send_files(attachments=atts, body=text or None, **kwargs)

    def get_extensible_media(self, media: str) -> tuple[str, list[str]]:
        try:
            media_dict = json.loads(media)
        except Exception as e:
            self.session.log.warning("Could not decipher extensible media: %r", e)
            raise ValueError

        text = ""
        files = []
        for item in media_dict.values():
            story = item.get("story_attachment")
            if not story:
                continue
            url = story.get("url")
            if not url:
                continue

            text += f"\n{url}"

            if title := story.get("title"):
                text += f" - {title}"

            if desc := story.get("description"):
                if desc_text := desc.get("text"):
                    text += f" - {desc_text}"

            if attachment := story.get("media"):
                if image := attachment.get("image"):
                    if url := image.get("uri"):
                        files.append(url)

        if not text and not files:
            raise ValueError

        return text, files

    async def get_attachment_url(
        self, attachment: mqtt_t.Attachment, thread_key, msg_id
    ):
        try:
            if v := attachment.video_info:
                return v.download_url
            if a := attachment.audio_info:
                return a.url
            if i := attachment.image_info:
                return i.uri_map.get(0)
        except AttributeError:
            media_id = getattr(attachment, "media_id", None)
            if media_id:
                return await self.session.api.get_file_url(
                    thread_key.thread_fbid or thread_key.other_user_id,
                    msg_id,
                    media_id,
                )

    async def send_fb_sticker(self, sticker_id: int, legacy_msg_id: str, **kwargs):
        resp = await self.session.api.fetch_stickers([sticker_id])
        await self.send_file(
            file_url=resp.nodes[0].preview_image.uri,
            legacy_file_id=f"sticker-{sticker_id}",
            legacy_msg_id=legacy_msg_id,
            **kwargs,
        )
